#include <htc.h>
#define _XTAL_FREQ  19660800     // oscillator frequency for _delay()


/*************************************************************************
I2C/TWI FUNCTIONS - I2C/TWI FUNCTIONS - I2C/TWI FUNCTIONS - I2C/TWI FUNCTIONS 
**************************************************************************/



//Initialise PIC16F887 for Hardware I2C operation
void I2C_INIT()
{
	SSPSTAT |=0b10000000; 		//_st_bit (SSPSTAT,SMP),Slew Rate Control Disabled
	SSPSTAT |=0b01000000;		//_cr_bit (SSPSTAT,CKE),Disable SMBus specific inputs
	SSPCON = 0x28;				//Setup I2C into Master Mode
	SSPADD = 98;				//Set the Baud Rate
	SSPCON2 = 0x00;				//Clear the control bits
	GIE = 1; 					//_st_bit(INTCON, GIE);
	TRISC = TRISC|0b00001000; 	//Configure SCL as Input
	TRISC = TRISC|0b00010000;	//Configure SDA as Input
}

// send a start bit sequence
void I2C_START()
{
	PIR1 &=~0b00001000; 		//cr_bit(PIR1,SSPIF),Clear SSP interrupt flag
	SSPCON2 |= 0x01;			//st_bit(SSPCON2,SEN),Initiate start condition
	while(SSPCON2 & 0x01);		// - as long as SEN is High - Wait for start bit to be generated
}

// send a repeated start bit sequence
void I2C_RESTART()
{
	PIR1 &=~0b00001000; 		//cr_bit(PIR1,SSPIF),Clear SSP interrupt flag
	SSPCON2 |= 0x02;			//st_bit(SSPCON2,RSEN),Initiate restart condition
	while(SSPCON2 & 0x02);		//	while(ts_bit(SSPCON2,RSEN)),Wait for restart bit to be generated
}

// send a stop bit sequence
void I2C_STOP()
{
	PIR1 &=~0b00001000; 		//cr_bit(PIR1,SSPIF),Clear SSP interrupt flag
	SSPCON2 |= 0x04;			//st_bit(SSPCON2,PEN),Initiate stop condition
	while(SSPCON2 & 0x02);		//while(ts_bit(SSPCON2,PEN)),Wait for stop bit to be generated
	__delay_ms(10);				//Wait before reusing the I2C BUS
}

// transmit a byte (this can be a devide adress - an index or just data - of 8 bit
// if return = 1 : No ack bit was received from the slave - if return = 0 = a correct acknowledge bit was received
char I2C_TRANSMIT_BYTE(char Data)
{
	PIR1 &=~0b00001000; 				//cr_bit(PIR1,SSPIF),Clear SSP interrupt flag
	SSPBUF=Data;						//Send byte
		while((PIR1 & 0x08) == 0);		//Wait for control bit to be sent
			if(SSPCON2 & 0b01000000)	//Check Acknowledgement
				return (1);				//No Acknowledgement
				else return (0);		//Acknowledgement received
}

//receive a byte of data - requested by the master and sent by the slave
// returns a byte of incoming data
// if last = 1 - last byte was received - If last = 0 - the master expects further I2C_RECEIVE_BYTE instructions
char I2C_RECEIVE_BYTE(char Last)
{
	PIR1 &=~0b00001000; 				//cr_bit(PIR1,SSPIF);
	SSPCON2 |=0x08;						//st_bit(SSPCON2,RCEN),Initiate Read
	while((PIR1 & 0x08) == 0);			//while(ts_bit(PIR1,SSPIF) == 0),Wait for data read
		if (Last)
			SSPCON2 |=0b00100000; 		//st_bit(SSPCON2,ACKDT),Send Nack
		else SSPCON2 &=~0b00100000;		//cr_bit(SSPCON2,ACKDT),Send Ack
			SSPCON2 |=0b00010000; 		//st_bit(SSPCON2,ACKEN),Initiate Nack
	while(SSPCON2&0b00010000);			//(ts_bit(SSPCON2,ACKEN)) Wait for data read
	return(SSPBUF);						//Store incoming data
}

//combination of all the I2C instructions to send a byte of date to a specific 
// index in a certain device
void I2C_SEND_MESSAGE(char Device_ID, char AddrH, char Data)
{
		char RetVal = 0;
		I2C_START();					//Start transaction
		I2C_TRANSMIT_BYTE(Device_ID);	//Transmit Device Address
		I2C_TRANSMIT_BYTE(AddrH);		//Transmit Internal Address
		I2C_TRANSMIT_BYTE(Data);		//Send Data byte
		I2C_STOP();						//Stop Transaction
}

//combination of all the I2C instructions to receive a byte of date from a specific 
// index in a certain device
char I2C_RECEIVE_MESSAGE(char Device_ID, char AddrH)
{
	
		char RetVal = 0;

		I2C_START();						//Start transaction
		I2C_TRANSMIT_BYTE(Device_ID);		//Transmit Device Address
		I2C_TRANSMIT_BYTE(AddrH);			//Transmit Internal Address

		I2C_RESTART();						//Restart transaction

		Device_ID = Device_ID | 0x01;		//Change Device ID to read mode
		I2C_TRANSMIT_BYTE(Device_ID);		//Transmit Device Address
		RetVal = I2C_RECEIVE_BYTE(1);		//Read data at address
		I2C_STOP();							//Stop Transaction

		return RetVal;						//Return data

}

/*************************************************************************
LCD FUNCTIONS - LCD FUNCTIONS - LCD FUNCTIONS - LCD FUNCTIONS 
**************************************************************************/

//Set the Cursor to a position specified by the next byte, where 1 is the top left and 80 is the bottom right
void LCD03_SET_CURSOR(char FCL_CURSOR)
{
	I2C_START();
	I2C_TRANSMIT_BYTE(0xC6);
	I2C_TRANSMIT_BYTE(0);
	I2C_TRANSMIT_BYTE(2);
	I2C_TRANSMIT_BYTE(FCL_CURSOR);
	I2C_STOP();
}

//Clears the whole Screen
void LCD03_CLEAR_SCREEN()
{
	I2C_START();
	I2C_TRANSMIT_BYTE(0xC6);
	I2C_TRANSMIT_BYTE(0);
	I2C_TRANSMIT_BYTE(12);
	I2C_STOP();
}

// put backlight on
void LCD03_BACKLIGHT_ON()
{
	I2C_START();
	I2C_TRANSMIT_BYTE(0xC6);
	I2C_TRANSMIT_BYTE(0);
	I2C_TRANSMIT_BYTE(19);
	I2C_STOP();
}

//put backlight off
void LCD03_BACKLIGHT_OFF()
{
	I2C_START();
	I2C_TRANSMIT_BYTE(0xC6);
	I2C_TRANSMIT_BYTE(0);
	I2C_TRANSMIT_BYTE(20);
	I2C_STOP();
}

/*
call LCD_WRITE_STRING in main programas follows:
char String[20] = "Hello World";
LCD_WRITE_STRING(String);
also
LCD_WRITE_STRING("Hello World");*/
void LCD_WRITE_STRING(  const char *pString )
{
while ( *pString != 0 ) 
	{
	I2C_SEND_MESSAGE(0xC6,0,*pString);
	pString++;
	}
}

// Writes an integer value to the LCD (16bit)
void LCD03_WRITE_INT(int ii)
{
	unsigned char j,jj,jjj,jjjj,jjjjj;
	j=((ii/10000) % 10)+48;  						// ? x 10000
	jj=((ii/1000) % 10)+48;                       	// ? x 1000
	jjj=((ii/100) % 10)+48;                       	// ? x 100
	jjjj=((ii/10) % 10)+48;                       	// ? x 10
	jjjjj=(ii % 10)+48;                           	// ? x 1
	if (j==48)   I2C_SEND_MESSAGE(0xC6,0,32);       // space if number does not exist 
	else        I2C_SEND_MESSAGE(0xC6,0,j);			// Write to LCD  (48 is the ASCII offset) 

	if ((jj==48)&(j==48))   I2C_SEND_MESSAGE(0xC6,0,32);
	else I2C_SEND_MESSAGE(0xC6,0,jj);	//putch(48+jj); 
               
	if ((jjj==48)&(jj==48))   I2C_SEND_MESSAGE(0xC6,0,32);
	else I2C_SEND_MESSAGE(0xC6,0,jjj);
                     
	if ((jjjj==48)&(jjj==48))   I2C_SEND_MESSAGE(0xC6,0,32);
	else I2C_SEND_MESSAGE(0xC6,0,jjjj);

	I2C_SEND_MESSAGE(0xC6,0,jjjjj);	              
}

// Writes a BYTE value to the LCD (8bit)
void LCD03_WRITE_BYTE(char ii)
{
	unsigned char jj,jjj,jjjj;

	jj=((ii/100) % 10)+48;                       	// ? x 100
	jjj=((ii/10) % 10)+48;                       	// ? x 10
	jjjj=(ii % 10)+48;                           	// ? x 1

	if ((jj==48))   I2C_SEND_MESSAGE(0xC6,0,32);
	else I2C_SEND_MESSAGE(0xC6,0,jj);	//putch(48+jj); 
               
	if ((jjj==48)&(jj==48))   I2C_SEND_MESSAGE(0xC6,0,32);
	else I2C_SEND_MESSAGE(0xC6,0,jjj);	//putch(48+jjj);
                     
	I2C_SEND_MESSAGE(0xC6,0,jjjj);	//putch(48+jjjj);                
}

//Clears a single line (LIne 0-1-2-3)
void LCD03_CLEAR_LINE(char line) // line 0-1-2-3
{
char cursor;
cursor = line *20;
LCD03_SET_CURSOR(cursor);
LCD_WRITE_STRING("                    ");	//20 x space
}
/*************************************************************************
STRING FUNCTIONS - STRING FUNCTIONS - STRING FUNCTIONS - STRING FUNCTIONS 
**************************************************************************/

//unsigned char STR_LEN( const char *p );
//This function returns the length of the string p.
unsigned char STR_LEN( const char *p ){
   unsigned char n = 0;
   while( *p != '\0' ){
      p++;
      n++;
   }
   return n;	
}
//bit str_equal( const char *p, const char *q );
//This function returns whether the strings p and q are equal.
bit STR_EQUAL( const char *p, const char *q ){
   for(;;){
      if( *p != *q ){
         return 0;		
      } 
      if( *p == '\0' ){
      	return 1;
      }
      p++;
      q++;
   }
}

bit STR_PREFIX( const char *p, const char *q ){
   for(;;){
      if( *p == '\0' ){
      	return 1;
      }
      if( *p != *q ){
         return 0;		
      } 
      p++;
      q++;
   }
}
//This function copies up to n characters from string q to string p. S is the size of string n.
void STR_COPY( char *p, unsigned char s, char *q, unsigned char n ){
   for(;;){
      *p = '\0';
      if( s == 0 ) return;
      if( n == 0 ) return;
      if( *q == '\0' ) return;
      *p = *q;
      p++;
      q++;
      s--;
      n--;
   }	
}
//bit char_in( char c, const char * s );
//This functions returns whether character c appears in string s.
bit CHAR_IN( char c, const char * s ){
   for(;;){
      if( *s == '\0' ) return 0;
      if( *s == c ) return 1;
      s++;
   } 			
}
//bit STRING_IN( const char * p, const char * q );
//This function returns whether the string p appears somewhere in string q.
bit STRING_IN( const char * p, const char * q ){
   for(;;){
   	  if( STR_PREFIX( p, q )) return 1;
      if( *q == '\0' ) return 0;
      q++;
   } 			
}

/*************************************************************************
RS232/UART FUNCTIONS - RS232/UART FUNCTIONS - RS232/UART FUNCTIONS 
**************************************************************************/


// BRGH=1, BRG16=1
#define UART_INIT( spbrg )    \
   TRISC6 = 0;		          \
   TRISC7 = 1;                \
   BAUDCTL = 0x08;            \
   SPBRGH = ((unsigned int)spbrg) >> 8;     \
   SPBRG = ((unsigned int)spbrg) & 0xFF;    \
   RCSTA = 0x90;              \
   TXSTA = 0x24;              \


// These functions initialize the UART at different baudrates.
// settings for Proton - PIC - 16F887 - Xtal 19660800
void UART_INIT_1200( void ){ UART_INIT( 4095 ); }
void UART_INIT_2400( void ){ UART_INIT( 2047 ); }
void UART_INIT_4800( void ){ UART_INIT( 1023 ); }
void UART_INIT_9600( void ){ UART_INIT( 511 ); }
void UART_INIT_19200( void ){ UART_INIT( 255 ); }
void UART_INIT_57600( void ){ UART_INIT( 84 ); }
void UART_INIT_115200( void ){ UART_INIT( 42 ); }
void UART_INIT_250000( void ){ UART_INIT( 19 ); }

//This function returns true when a received character is available, false otherwise.
bit UART_CHAR_AVAILABLE( void ){
   return RCIF;
}

//This function waits for a character to become available, reads it and returns it. Note that this function will
//not return until a character is available. If you want to avoid such waiting, use the function
//uart_char_available to check whether a character is available, and call uart_char_receive only when one is
//available.
char UART_CHAR_RECEIVE( void ){
	 while(!RCIF) continue;
   return RCREG;
}

//This returns true when the transmitter is idle- and can not be used
bit UART_SEND_IDLE( void ){
   return TXIF;
}

//This function waits for the transmitter to become idle, and then starts transmission of the character. It
//does not wait for the transmission to complete.
void UART_CHAR_SEND( char c ){
   while( !TXIF ) continue;
	 TXREG = c;
}

//This function uses uart_char_send to send a 0-terminated string of characters.
void UART_STRING_SEND( const char *p ){
    while( *p != 0 ){
       UART_CHAR_SEND( *p );
       p++;
    }
}

//This function receives a string of characters. The pointer received must point to a caller-allocated string.
//Characters that are in the string ignore will be skipped. Receiving will terminate as soon as one of the
//following conditions holds:
//		? received_size - 1 characters have been received (not counting skipped characters, - 1 to have room
//			for the terminating \0), or
//		? milliseconds milliseconds have elapsed, or
//		? A character that is in the string terminate has been received.

void UART_STRING_RECEIVE( 
   char *received,
   unsigned char received_size, 
   unsigned int milliseconds, 
   const char * terminate,
   const char * ignore 
){
   unsigned char i, t;
   char c;   
   t = 0;	 
   i = 0;
   for(;;){
   	  received[ i ] = '\0';
      if( i + 1 == received_size ){
         return;
      }
      if( UART_CHAR_AVAILABLE() ){
      	 c = UART_CHAR_RECEIVE();
         if( ! CHAR_IN( c, ignore )){
            received[ i ] = c;
            i++;
         }
         if( CHAR_IN( c, terminate )){
            return;
         }
      }	else {
   	     __delay_us( 100 );
         t++;
         if( t == 10 ){
   	        t = 0;
            milliseconds--;
            if( milliseconds == 0 ){
               return;
            }
         }
      }   
   }   	
}	

/*************************************************************************
ADC FUNCTIONS - ADC FUNCTIONS - ADC FUNCTIONS - ADC FUNCTIONS 
**************************************************************************/

// returns a 10 bit result from an AD conversion from a certain ADC channel
//Channels 0-13
// A 10 bit result needs to be loaded in an integer variable
unsigned int READ_ADC_INT_CHANNEL( unsigned char channel ){ 
	
	 // Fosc/32; channel; ADON 
   ADCON0 = 0x41 + ( channel << 2 ); 
   
   // right justified; Vss/Vdd ref
   ADCON1 = 0x80; 
   
   // configure the pin as analog
   if( channel < 8 ){ 
      ANSEL |= ( 1 << channel );
   } else if( channel < 14 ){
      ANSELH |= ( 1 << ( channel - 8 ));	
   }
   
   // configure the pin as input
   switch( channel ){
      case  0 : TRISA0 = 1; break;
      case  1 : TRISA1 = 1; break;
      case  2 : TRISA2 = 1; break;
      case  3 : TRISA3 = 1; break;
      case  4 : TRISA5 = 1; break;
      case  5 : TRISE0 = 1; break;
      case  6 : TRISE1 = 1; break;
      case  7 : TRISE2 = 1; break;
      case  8 : TRISB2 = 1; break;
      case  9 : TRISB3 = 1; break;
      case 10 : TRISB1 = 1; break;
      case 11 : TRISB4 = 1; break;
      case 12 : TRISB0 = 1; break;
      case 13 : TRISB5 = 1; break;
      default: break;
   }
   
   // p107 requires 4.67 us
   __delay_us( 6 );
   
   // start conversion and wait for completion
   ADCON0 |= 2;
   while( ADCON0 & 2 ){}
   
   // return A/D conversion result
   return ( ((unsigned int) ADRESH ) << 8 ) + ADRESL;	
}

// returns a 8 bit result from an AD conversion from a certain ADC channel
//Channels 0-13
// A 8 bit result needs to be loaded in a BYTE variable
unsigned char READ_ADC_BYTE_CHANNEL( unsigned char channel ){ 
	
	 // Fosc/32; channel; ADON 
   ADCON0 = 0x41 + ( channel << 2 ); 
   
   // right justified; Vss/Vdd ref
   ADCON1 = 0x00; 
   
   // configure the pin as analog
   if( channel < 8 ){ 
      ANSEL |= ( 1 << channel );
   } else if( channel < 14 ){
      ANSELH |= ( 1 << ( channel - 8 ));	
   }
   
   // configure the pin as input
   switch( channel ){
      case  0 : TRISA0 = 1; break;
      case  1 : TRISA1 = 1; break;
      case  2 : TRISA2 = 1; break;
      case  3 : TRISA3 = 1; break;
      case  4 : TRISA5 = 1; break;
      case  5 : TRISE0 = 1; break;
      case  6 : TRISE1 = 1; break;
      case  7 : TRISE2 = 1; break;
      case  8 : TRISB2 = 1; break;
      case  9 : TRISB3 = 1; break;
      case 10 : TRISB1 = 1; break;
      case 11 : TRISB4 = 1; break;
      case 12 : TRISB0 = 1; break;
      case 13 : TRISB5 = 1; break;
      default: break;
   }
   
   // p107 requires 4.67 us
   __delay_us( 6 );
   
   // start conversion and wait for completion
   ADCON0 |= 2;
   while( ADCON0 & 2 ){}
   
   // return A/D conversion result
   return ADRESH;	
}